#!/usr/bin/python
# -*- coding:utf-8 -*-
import os, sys
import subprocess
from struct import pack
from string import upper, split
from time import strftime
from re import search

#--------------------------------------------------------------------------------------------------------
# ABOUT THE PROGRAM

# mosaique_aster_gdem.py - Achieves a mosaic of ASTER GDEM tiles
# Copyright (C) 2015 Arthur Delorme
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# (contact: delorme@ipgp.fr)

# Call from a terminal: ./mosaique_aster_gdem.py arg1 arg2 arg3
#    with arg1: ASTER GDEM V1 or V2 (= 1 or 2)
#         arg2: mosaic's top-left corner (ex: N02W002)
#         arg3: mosaic's bottom-right corner (ex: S02E002)
#
# By default, ASTER GDEM tiles (ASTGTM_XXXXXXX_dem.tif or ASTGTM2_XXXXXXX_dem.tif) must be stored in
# the current directory, and the output will be generated there too.
# You can modify these parameters by editing chemin_dalles_aster and output_dir in the program.
#
# Notes:
#    - the name of the ASTER GDEM tiles consists of the coordinates of the bottom-left pixel's center

#--------------------------------------------------------------------------------------------------------
# CONSTANTES --> A EDITER SI NECESSAIRE (LE REPERTOIRE COURANT EST LE CHOIX PAR DEFAUT)

chemin_dalles_aster = '.' # Répertoire qui contient les dalles ASTER GDEM (ASTGTM_XXXXXXX_dem.tif ou ASTGTM2_XXXXXXX_dem.tif)
output_dir = '.' # Répertoire où sont stockés les fichiers de sortie

#--------------------------------------------------------------------------------------------------------
# TESTS SUR LES CONSTANTES

if not chemin_dalles_aster:
    chemin_dalles_aster = '.'
if not output_dir:
    output_dir = '.'

#--------------------------------------------------------------------------------------------------------
# TESTS SUR LES ENTREES

# Vérification du nombre d'arguments
if len(sys.argv) == 4:
    version = sys.argv[1]
    liste_coord = [upper(sys.argv[2]),upper(sys.argv[3])]
    # Vérification de la version
    if not version in ('1', '2'):
        sys.exit("Error: version number must be 1 or 2.\nRun ./mosaique_aster_gdem.py -h for more informations.")
    else:
        if version == '1':
            prefixe = 'ASTGTM_'
            nom_donnees = 'ASTER GDEM Elevation'
        else:
            prefixe = 'ASTGTM2_'
            nom_donnees = 'ASTER GDEM Elevation Version 2'
    # Vérification du format
    format = True # Booléen
    for liste_arg in liste_coord:
        if not search('^[NS][0-9]{2}[EOW][0-9]{3}$', liste_arg):
            format = False
    # Vérification des valeurs
    if format:
        valeurs_ok = True
        for liste_arg in liste_coord:
            lat = int(liste_arg[1:3])
            lon = int(liste_arg[4:7])
            if (lat > 90 or lon > 180):
                valeurs_ok = False
        if not valeurs_ok:
            sys.exit("Error: the values of the coordinates are absurd.\nRun ./mosaique_aster_gdem.py -h for more informations.")
    else:
        sys.exit("Error: coordinates must be in the form 'N01E001'.\nRun ./mosaique_aster_gdem.py -h for more informations.")
else:
    print """
    ---------------------------------------------------------------------------------
   | mosaique_aster_gdem.py - Copyright (C) 2015 Arthur Delorme                      |
   |---------------------------------------------------------------------------------|
   | Achieves a mosaic of ASTER GDEM tiles                                           |
   | (contact: delorme@ipgp.fr)                                                      |
   |                                                                                 |
   | This program is distributed under the terms of the GNU General Public License.  |
   | This program comes with ABSOLUTELY NO WARRANTY. This is free software, and you  |
   | are welcome to redistribute it under certain conditions.                        |
   | (see license.txt for details)                                                   |
   |                                                                                 |
   | Call from a terminal: ./mosaique_aster_gdem.py arg0 arg1 arg2                   |
   |    with arg0: ASTER GDEM V1 or V2 (= 1 or 2)                                    |
   |         arg1: mosaic's top-left corner (ex: N02W002)                            |
   |         arg2: mosaic's bottom-right corner (ex: S02E002)                        |
   |                                                                                 |
   | By default, ASTER GDEM tiles (ASTGTM_XXXXXXX_dem.tif or                         |
   | ASTGTM2_XXXXXXX_dem.tif) must be stored in the current directory, and the       |
   | output will be generated there too.                                             |
   | You can modify these parameters by editing chemin_dalles_aster and output_dir   |
   | in the program.                                                                 |
   |                                                                                 |
   | Notes:                                                                          |
   |    - the name of the ASTER GDEM tiles consists of the coordinates of the        |
   |      bottom-left pixel's center                                                 |
    ---------------------------------------------------------------------------------
"""
    sys.exit()

#--------------------------------------------------------------------------------------------------------
# CHOIX DES DALLES

lat_min = int(liste_coord[1][1:3])
if liste_coord[1][0] == 'S':
    lat_min *= -1

lat_max = int(liste_coord[0][1:3])
if liste_coord[0][0] == 'S':
    lat_max *= -1

lon_min = int(liste_coord[0][4:7])
if liste_coord[0][3] in ('W', 'O'):
    lon_min *= -1

lon_max = int(liste_coord[1][4:7])
if liste_coord[1][3] in ('W', 'O'):
    lon_max *= -1

delta_lat = lat_max-lat_min
delta_lon = lon_max-lon_min

# Si le coin haut-gauche de la mosaïque n'est pas strictement en haut à gauche du coin bas-droit
if delta_lat < 1 or delta_lon < 1:
    sys.exit("Error: the top-left corner is not located at the top-left from the bottom-right corner.\nRun ./mosaique_aster_gdem.py -h for more informations.")

lat_max -= 1 # On veut le coin bas-gauche de la dalle en haut à gauche
lon_max -= 1 # On veut le coin bas-gauche de la dalle en bas à droite

liste_lat = []
i = 0
lat = lat_max
while 1:
    if i == delta_lat:
        break
    liste_lat.append([])
    j = 0
    if lat >= 0:
        lat_str = '{:0>2d}'.format(lat)
    else:
        lat_tmp = lat*-1
        lat_str = '{:0>2d}'.format(lat_tmp)
    while 1:
        if j == delta_lon:
            break
        if lat >= 0:
            liste_lat[i].append('N' + lat_str)
        else:
            liste_lat[i].append('S' + lat_str)
        j += 1
    lat = lat-1
    i += 1

liste_lon = []
i = 0
lon = lon_min
while 1:
    if i == delta_lon:
        break
    if lon >= 0:
        lon_str = '{:0>3d}'.format(lon)
        liste_lon.append('E' + lon_str)
    else:
        lon_tmp = lon*-1
        lon_str = '{:0>3d}'.format(lon_tmp)
        liste_lon.append('W' + lon_str)
    lon = lon+1
    i += 1

liste_fic = []
i = 0
for liste in liste_lat:
    j = 0
    liste_fic.append([])
    for lat_str in liste:
        lat_lon = '{0}{1}{2}_dem.tif'.format(prefixe, lat_str, liste_lon[j])
        liste_fic[i].append(lat_lon)
        j += 1
    i += 1

#--------------------------------------------------------------------------------------------------------
# TEST SUR LA LISTE DE DALLES --> RESTRICTION DE LA SELECTION

aucun_fic_tot = True # Booléen : vrai si aucune des dalles n'est dans la base (sélection vide)

#-------------------
# Test sur chaque ligne de dalles (pas sur les colonnes)

result_lig = [] # Liste qui accueille les résultat du test sur chaque ligne
# Test ligne par ligne
for liste in liste_fic:
    aucun_fic_ligne = True # Booléen : vrai si aucune des dalles de la ligne n'est dans la base (ligne vide)
    for f in liste:
        if os.path.exists(os.path.join(chemin_dalles_aster, f)): # Si le fichier ASTER GDEM existe
            aucun_fic_ligne = False
            aucun_fic_tot = False
    result_lig.append(aucun_fic_ligne)

# Modification de la liste des résultats : à partir du moment où une ligne contient au moins une dalle, les lignes suivantes
# ne sont pas supprimées, même si elles sont vides (elles seront comblées avec des zéros)
result_lig_modif = [] # Liste modifiée
stop = False # Booléen : vrai à partir du moment où une ligne contient au moins une dalle
for result in result_lig:
    if stop: # Si une ligne non-vide a été détectée, les suivantes ne sont pas supprimées
        result_lig_modif.append(0)
    else: # Sinon, on recopie la liste et on contrôle la valeur
        result_lig_modif.append(result)
        if result == 0:
            stop = True

result_lig.reverse() # On inverse la liste initiale pour détecter les lignes à supprimer à la fin
result_lig_modif2 = [] # Liste inverse modifiée
stop = False # Booléen : vrai à partir du moment où une ligne contient au moins une dalle
for result in result_lig:
    if stop: # Si une ligne non-vide a été détectée, les suivantes ne sont pas supprimées
        result_lig_modif2.append(0)
    else: # Sinon, on recopie la liste et on contrôle la valeur
        result_lig_modif2.append(result)
        if result == 0:
            stop = True
result_lig_modif2.reverse() # On inverse la liste modifiée pour la remettre à l'endroit

# Fusion des 2 listes modifiées
result_lig = []
i = 0
for result in result_lig_modif:
    if result == 1 or result_lig_modif2[i] == 1:
        result_lig.append(1)
    else:
        result_lig.append(0)
    i += 1

# Test sur la liste pour prévenir l'utilisateur si des lignes vont être supprimées
lignes_suppr = False # Booléen : vrai si des lignes vont être supprimées
for val in result_lig:
    if val == 1:
        lignes_suppr = True
        break

# Suppression des lignes vides
liste_fic_tmp = []
i = 0
for liste in liste_fic:
    if result_lig[i] == 0:
        liste_fic_tmp.append(liste_fic[i])
    i += 1

liste_fic_tmp_transp = map(list, zip(*liste_fic_tmp)) # Transposition de la liste pour effectuer le traitement par ligne plutôt que par colonne

# Si aucune dalle n'est dans la base (sélection vide)
if aucun_fic_tot:
    sys.exit("Error: the selection contains no data!\nMake sure that all the necessary ASTER GDEM tiles are present in the directory '{0}'.".format(chemin_dalles_aster))

#-------------------
# Test sur chaque colonne de dalles

result_lig = [] # Liste qui accueille les résultat du test sur chaque ligne
# Test ligne par ligne
for liste in liste_fic_tmp_transp:
    aucun_fic_ligne = 1 # Booléen : vrai si aucune des dalles de la ligne n'est dans la base (ligne vide)
    for f in liste:
        if os.path.exists(os.path.join(chemin_dalles_aster, f)): # Si le fichier ASTER GDEM existe
            aucun_fic_ligne = 0
    result_lig.append(aucun_fic_ligne)

# Modification de la liste des résultats : à partir du moment où une ligne contient au moins une dalle, les lignes suivantes
# ne sont pas supprimées, même si elles sont vides (elles seront comblées avec des zéros)
result_lig_modif = [] # Liste modifiée
stop = False # Booléen : vrai à partir du moment où une ligne contient au moins une dalle
for result in result_lig:
    if stop: # Si une ligne non-vide a été détectée, les suivantes ne sont pas supprimées
        result_lig_modif.append(0)
    else: # Sinon, on recopie la liste et on contrôle la valeur
        result_lig_modif.append(result)
        if result == 0:
            stop = True

result_lig.reverse() # On inverse la liste initiale pour détecter les lignes à supprimer à la fin
result_lig_modif2 = [] # Liste inverse modifiée
stop = False # Booléen : vrai à partir du moment où une ligne contient au moins une dalle
for result in result_lig:
    if stop: # Si une ligne non-vide a été détectée, les suivantes ne sont pas supprimées
        result_lig_modif2.append(0)
    else: # Sinon, on recopie la liste et on contrôle la valeur
        result_lig_modif2.append(result)
        if result == 0:
            stop = True
result_lig_modif2.reverse() # On inverse la liste modifiée pour la remettre à l'endroit

# Fusion des 2 listes modifiées
result_lig = []
i = 0
for result in result_lig_modif:
    if result == 1 or result_lig_modif2[i] == 1:
        result_lig.append(1)
    else:
        result_lig.append(0)
    i += 1

# Test sur la liste pour prévenir l'utilisateur si des lignes vont être supprimées
if not lignes_suppr:
    for val in result_lig:
        if val == 1:
            lignes_suppr = True
            break

# Suppression des lignes vides
liste_fic_tmp = []
i = 0
for liste in liste_fic_tmp_transp:
    if result_lig[i] == 0:
        liste_fic_tmp.append(liste_fic_tmp_transp[i])
    i += 1

if lignes_suppr:
    print "Warning: some ASTER GDEM files are missing: the mosaic's limits have been reduced accordingly."

liste_fic = map(list, zip(*liste_fic_tmp)) # On retranspose pour remettre la liste d'aplomb

#--------------------------------------------------------------------------------------------------------
# ASSEMBLAGE DES DALLES

nom_fic_out = '{0}_{1}.dem'.format(liste_coord[0], liste_coord[1])
chem_fic_out = os.path.join(output_dir, nom_fic_out)

# Teste si un fichier portant le même nom existe déjà et propose un autre nom si nécessaire
term_fic = 0
while 1:
    if os.path.isfile(chem_fic_out):
        term_fic += 1
        nom_fic_out = '{0}_{1}_{2}.dem'.format(liste_coord[0], liste_coord[1], term_fic)
        chem_fic_out = os.path.join(output_dir, nom_fic_out)
    else:
        break

# Ouverture du fichier de sortie
try:
    fic_out = open(chem_fic_out,'wb')
except:
    sys.exit("Error: cannot write in the file '{0}'!\nMake sure that the directory '{1}' exists and that you are allowed to write in it.".format(chem_fic_out, output_dir))

# Traitement des données
i = 0
manque_dalle = False # Booléen : vrai si au moins un fichier ASTER n'existe pas (comblé par des zéros)
for liste in liste_fic:
    # Ouverture des fichiers, ligne par ligne
    ligne = [] # Liste contenant les variables pour ouvrir les fichiers d'une ligne complète (une valeur de latitude)
    for nom_fic in liste:
        if os.path.exists(os.path.join(chemin_dalles_aster, nom_fic)): # Si le fichier ASTER GDEM existe
            try:
                ligne.append(open(os.path.join(output_dir, nom_fic),'rb'))
            except:
                print "{0}: cannot read the file. Make sure that the file is present in the directory '{1}' and the you are allowed to read it.".format(nom_fic, output_dir)
                
                # Fermeture des fichiers de la ligne en cours
                for f in ligne:
                    f.close()
                
                fic_out.close()
                os.remove(chem_fic_out)
                
                sys.exit()
        
        else: # Si le fichier ASTER n'existe pas, on indique qu'il faut combler avec des zéros
            ligne.append('zeros')
            manque_dalle = True
    
    # Assemblage de la ligne
    # Entier signé 16 bits --> 2 octets
    size = 3601
    j = 0
    while 1:
        if j == size: # Quand il n'y a plus de lignes
            break
        k = 0
        for fic in ligne:
            if j == 0 and fic != 'zeros':
                fic.read(8) # On ne prend pas en compte les 4 premières valeurs qui ne sont pas des valeurs d'altitude (ainsi que les valeurs strictement supérieures à 12 967 205 = 3601*3601+4)
            if k != len(ligne) - 1: # Pour chaque dalle de la ligne, sauf la dernière
                if fic == 'zeros':
                    # Encode '0' sur 2 octets en little endian
                    data = pack('<h',0)*(size-1) # 1 ligne = 3601 samples - 1 pour retirer la colonne qui recouvre la dalle suivante
                else:
                    data = fic.read((size-1)*2) # 1 ligne = 3601 samples - 1 pour retirer la colonne qui recouvre la dalle suivante
                if fic != 'zeros':
                    fic.read(2) # Ignore la dernière colonne
            
            else: # Pour la dernière dalle de la ligne
                if fic == 'zeros':
                    # Encode '0' sur 2 octets en little endian
                    data = pack('<h',0)*size # 1 ligne = 3601 samples
                else:
                    data = fic.read(size*2) # 1 ligne = 3601 samples
                
            if (i == 0) or (i != 0 and j > 0): # Si c'est la 1ère ligne de la mosaïque, on prend en compte la 1ère ligne des images, sinon non (recouvrement avec la ligne du dessus)
                fic_out.write(data)
            k += 1
        j += 1
    
    # Fermeture des fichiers de la ligne
    for f in ligne:
        if f != 'zeros':
            f.close()
    
    i += 1

# Fermeture du fichier de sortie
fic_out.close()

if manque_dalle:
    print "Warning: some ASTER GDEM files are missing: missing data have been replaced by zeros."
print "--> " + chem_fic_out

#--------------------------------------------------------------------------------------------------------
# FICHIER HEADER (.hdr)

nb_lig = str(size+(len(liste_fic)-1)*(size-1))
nb_col = str(size+(len(liste_fic[0])-1)*(size-1))

coord_hg = split(liste_fic[0][0], '_')

# Latitude du coin haut-gauche en secondes d'arc
if coord_hg[1][0] == 'N':
    lat_float = (int(coord_hg[1][1:3])+1)*3600+0.5
elif coord_hg[1][0] == 'S':
    lat_float = (int(coord_hg[1][1:3])-1)*3600-0.5
    lat_float *= -1

# Longitude du coin haut-gauche en secondes d'arc
if coord_hg[1][3] == 'E':
    lon_float = int(coord_hg[1][4:7])*3600-0.5
elif coord_hg[1][3] == 'W':
    lon_float = int(coord_hg[1][4:7])*3600+0.5
    lon_float *= -1

# Ouverture du fichier header
nom_fic_hdr = nom_fic_out+'.hdr'
chem_fic_hdr = os.path.join(output_dir, nom_fic_hdr)

try:
    fic_hdr = open(chem_fic_hdr,'w')
except:
    os.remove(chem_fic_out)
    sys.exit("Cannot write in the file '{0}'!\nMake sure that the directory '{1}' exists and that you are allowed to write in it.".format(chem_fic_hdr, output_dir))

# Écriture dans le fichier header
fic_hdr.write("ENVI\ndescription = {{\n  ENVI File, Created [{0}]}}\n"
    "samples = {1}\nlines   = {2}\nbands   = 1\nheader offset = 0\n"
    "file type = ENVI Standard\ndata type = 2\ninterleave = bsq\nsensor type = Unknown\nbyte order = 0\n"
    "map info = {{Geographic Lat/Lon, 1.0000, 1.0000, {3}, {4}, 1.0000000000e+00,"
    "1.0000000000e+00, WGS-84, units=Seconds}}\nwavelength units = Unknown\nband names = {{\n"
    " {5}}}".format(strftime("%a %b %d %H:%M:%S %Y"), nb_col, nb_lig, lon_float, lat_float, nom_donnees))

fic_hdr.close()

print "--> {0}\nThe mosaic were successfully generated!".format(chem_fic_hdr)
